"""Main module."""
import numpy as np
import numba as nb
import xarray as xr
import xsimlab as xs
import pandas as pd
import math
import fastscape
from fastscape.processes.erosion import TotalErosion
from fastscape.processes.channel import DifferentialStreamPowerChannelTD
from fastscape.processes.hillslope import DifferentialLinearDiffusion
from fastscape.processes.grid import UniformRectilinearGrid2D
import xsimlab as xs

from fastscape.processes.boundary import BorderBoundary
from fastscape.processes.channel import (StreamPowerChannel,
                                 DifferentialStreamPowerChannelTD)
from fastscape.processes.context import FastscapelibContext
from fastscape.processes.flow import DrainageArea, SingleFlowRouter, MultipleFlowRouter, FlowRouter, FlowAccumulator
from fastscape.processes.erosion import TotalErosion
from fastscape.processes.grid import RasterGrid2D
from fastscape.processes.hillslope import LinearDiffusion, DifferentialLinearDiffusion
from fastscape.processes.initial import (BareRockSurface,
                                 Escarpment,
                                 FlatSurface,
                                 NoErosionHistory)
from fastscape.processes.main import (Bedrock,
                              StratigraphicHorizons,
                              SurfaceTopography,
                              SurfaceToErode,
                              TerrainDerivatives,
                              TotalVerticalMotion,
                              UniformSedimentLayer)
from fastscape.processes.marine import MarineSedimentTransport, Sea
from fastscape.processes.tectonics import (BlockUplift,
                                   SurfaceAfterTectonics,
                                   TectonicForcing,
                                   TwoBlocksUplift)
from ._looper import iterasator
import fastscape_litho._helper as fhl


##############################################################
##############################################################
####################### Topographic Analysis #################
##############################################################
##############################################################



@xs.process
class QuickTA:
  """
     This Process class provides quick runtime topographic analysis. It analyses the landscapes As if it was a "real" DEM.
     Authors: B.G
  """
    
  # UniformRectilinearGrid2D
  dx = xs.foreign(UniformRectilinearGrid2D, 'dx')
  dy = xs.foreign(UniformRectilinearGrid2D, 'dy')

  nx = xs.foreign(UniformRectilinearGrid2D, 'nx')
  ny = xs.foreign(UniformRectilinearGrid2D, 'ny')

  x = xs.foreign(UniformRectilinearGrid2D, 'x')
  y = xs.foreign(UniformRectilinearGrid2D, 'y')

  fs_context = xs.foreign(FastscapelibContext, 'context')

  # FlowRouter
  elevation = xs.foreign(SurfaceToErode, 'elevation')
  stack = xs.foreign(FlowRouter, 'stack')
  nb_receivers = xs.foreign(FlowRouter, 'nb_receivers')
  receivers = xs.foreign(FlowRouter, 'receivers')
  lengths = xs.foreign(FlowRouter, 'lengths')
  weights = xs.foreign(FlowRouter, 'weights')
  nb_donors = xs.foreign(FlowRouter, 'nb_donors')
  donors = xs.foreign(FlowRouter, 'donors')
  boundary_conditions = xs.foreign(BorderBoundary, "status")
  erosion_rate = xs.foreign(TotalErosion, "rate")

  flowacc = xs.foreign(FlowAccumulator, 'flowacc')

  theta_chi = xs.variable(intent = 'in', description = "Theta for chi calculation")
  A_0_chi = xs.variable(intent = 'in', description = "Reference drainage area for chi calculations")
  minAcc = xs.variable(intent = 'in', description = "minimum A/Q to do the analysis")
  output_prefix = xs.variable(intent = 'in', default = "TA_output")
  main_divide_contrast_distance = xs.variable(intent = 'in', default = 5000)
  thetas_default = np.arange(0.05,1,0.05)
  # main_divide_cross_theta_list = xs.variable(intent = 'in', dims= ((),'n_theta'), default = 5)

  internal_chi = xs.variable(intent = 'out',
    dims=('nodes'),
    description='chi coordinate'
    )

  chi = xs.on_demand(
    dims=('y', 'x'),
    description='chi coordinate'
    )

  ksn = xs.on_demand(
    dims=('y', 'x'),
    description='Local k_sn index (dz/dchi)'
    )

  ksnSA = xs.on_demand(
    dims=('y', 'x'),
    description='Local k_sn index (ksn = S A ^ theta)'
  )

  ksnSA_cross_theta = xs.on_demand(
    dims=('n_theta'),
    description='Local k_sn index (ksn = S A ^ theta)'
  )

  main_divide = xs.on_demand(
    dims=('y', 'x'),
    description = 'Label the areas draining to each fixed_value edges. Steepest Descent approximation.'
  )

  main_divide_cross_theta = xs.on_demand(
    dims=('n_theta'),
    description = 'Label the areas draining to each fixed_value edges. Steepest Descent approximation.'
  )

  main_drainage_divides_migration_index = xs.on_demand(description = "Represents the main drainage divide variations at each time step")

  def initialize(self):
    # getting boundary conditions right for the iteratool
    self.node_type = np.ones((self.ny,self.nx), dtype = np.int16)
    #Was drainage divide is an intermediate checker
    self.was_DD = np.zeros((self.ny,self.nx), dtype = np.bool)

    #left, right, top, bottom boundary conditions
    if(self.boundary_conditions[0] == "fixed_value"):
      self.node_type[:,0] = 0
    else:
      self.node_type[:,0] = 2
    if(self.boundary_conditions[1] == "fixed_value"):
      self.node_type[:,-1] = 0
    else:
      self.node_type[:,-1] = 2
    if(self.boundary_conditions[2] == "fixed_value"):
      self.node_type[0,:] = 0
    else:
      self.node_type[0,:] = 2
    if(self.boundary_conditions[3] == "fixed_value"):
      self.node_type[-1,:] = 0
    else:
      self.node_type[-1,:] = 2

    # getting the iteratoolto work
    self.iteratool = iterasator(self.node_type.ravel(),self.nx,self.ny, self.dx, self.dy)


  @xs.runtime(args = ["step_end", "step"])
  def run_step(self,step_end,step):
    self.step_end = str(step_end)
    self.step = str(step)

    # Just checking if self.is_multiple_flow
    self.is_multiple_flow = True if self.receivers.ndim > 1 else False
    if(self.is_multiple_flow == False):
      self.internal_chi = fhl.chiculation_SF(self.stack, self.receivers, self.nb_donors, self.donors, self.lengths, 
        self.elevation.ravel(), self.flowacc.ravel(), self.A_0_chi, self.theta_chi, self.minAcc).reshape(self.ny,self.nx)
    else:
      self.internal_chi = fhl.chiculation_MF(self.stack, self.receivers, self.nb_receivers, self.lengths, self.weights, 
        self.elevation, self.flowacc.ravel(), self.A_0_chi, self.theta_chi, self.minAcc)

  def finalize_step(self):

    basins = fhl.basination_SF(self.fs_context["stack"].astype('int') - 1, self.fs_context["rec"].astype('int') - 1)
    self.was_DD = fhl.is_draiange_divide_SF(basins, self.iteratool).reshape(self.ny, self.nx)




  def get_slope(self):
    if(self.is_multiple_flow):
      return fhl.slope_MF(self.elevation.ravel(), self.receivers, self.lengths, self.nb_receivers, self.weights)
    else:
      return fhl.slope_SF(self.elevation.ravel(), self.receivers, self.lengths)

  @chi.compute
  def _chi(self):
    return self.internal_chi.reshape(self.ny,self.nx)
    

  @ksn.compute
  def _ksn(self):
    if(self.is_multiple_flow == False):
      return fhl.ksn_calculation_SF(self.elevation.ravel(), self.internal_chi.ravel(), self.receivers, self.stack).reshape(self.ny,self.nx)
    else:
      return fhl.ksn_calculation_MF(self.elevation.ravel(), self.internal_chi.ravel(), self.receivers,self.nb_receivers, self.weights, self.stack).reshape(self.ny,self.nx)

  @ksnSA.compute
  def _ksnSA(self):
    ret = np.power(self.flowacc,self.theta_chi) * self.get_slope().reshape(self.ny,self.nx)
    ret[self.flowacc<self.minAcc] = 0
    return ret





  @main_drainage_divides_migration_index.compute
  def _main_drainage_divides_migration_index(self):
    sst = self.fs_context["stack"].astype('int') - 1
    # print(sst)
    # print(self.receivers)
    basins = fhl.basination_SF(sst, self.fs_context["rec"].astype('int') - 1)
    is_DD = fhl.is_draiange_divide_SF(basins, self.iteratool).reshape(self.ny, self.nx)
    return is_DD[(is_DD == 1) & (self.was_DD == 0) ].shape[0]


  @main_divide.compute
  def _main_divide(self):
    # First labelling the outlets
    main_divide = np.zeros_like(self.elevation,dtype = np.int16)
    #left, right, top, bottom boundary conditions
    if(self.boundary_conditions[0] == "fixed_value"):
      main_divide[:,0] = 1
    if(self.boundary_conditions[1] == "fixed_value"):
      main_divide[:,-1] = 2
    if(self.boundary_conditions[2] == "fixed_value"):
      main_divide[0,:] = 3
    if(self.boundary_conditions[3] == "fixed_value"):
      main_divide[-1,:] = 4

    main_divide = main_divide.ravel()

    fhl.extract_main_regions(self.fs_context["stack"].astype('int') - 1,
        main_divide, self.fs_context["rec"].astype('int') - 1)


    output = fhl.chi_contrast_across_divides(main_divide, self.internal_chi.ravel(), self.iteratool, self.main_divide_contrast_distance )
    df = pd.DataFrame({"X": output[0], "Y": output[1], "chi_med_zone": output[2], "chi_med_others": output[3],
      "chi_max_zone": output[4], "chi_max_others": output[5], "zone": output[6].astype(np.int)})
    
    tstep = str(self.step)
    while(len(tstep)<6):
      tstep = "0" +  tstep

    df.to_csv(self.output_prefix + "%s_mainDDinfo.csv"%(tstep), index = False)

    return main_divide.reshape(self.ny,self.nx)

  @main_divide_cross_theta.compute
  def _main_divide_cross_theta(self):

    dfs = []
    # First labelling the outlets
    main_divide = np.zeros_like(self.elevation,dtype = np.int16)
    #left, right, top, bottom boundary conditions
    if(self.boundary_conditions[0] == "fixed_value"):
      main_divide[:,0] = 1
    if(self.boundary_conditions[1] == "fixed_value"):
      main_divide[:,-1] = 2
    if(self.boundary_conditions[2] == "fixed_value"):
      main_divide[0,:] = 3
    if(self.boundary_conditions[3] == "fixed_value"):
      main_divide[-1,:] = 4

    main_divide = main_divide.ravel()

    fhl.extract_main_regions(self.fs_context["stack"].astype('int') - 1,
        main_divide, self.fs_context["rec"].astype('int') - 1)

    thetas = np.arange(0.05,1,0.05)
    dfs = []
    for theta in thetas:

      this_chi = fhl.chiculation_SF(self.fs_context["stack"].astype('int') - 1, 
        self.fs_context["rec"].astype('int') - 1, self.fs_context["ndon"].astype('int'),
         self.fs_context["don"].astype('int') - 1, self.fs_context["length"], 
          self.elevation.ravel(), self.flowacc.ravel(), self.A_0_chi, theta, self.minAcc)

      output = fhl.chi_contrast_across_divides(main_divide, this_chi.ravel(), self.iteratool, self.main_divide_contrast_distance )
      dfs.append(pd.DataFrame({"X": output[0], "Y": output[1], "chi_med_zone": output[2], "chi_med_others": output[3],
        "chi_max_zone": output[4], "chi_max_others": output[5], "zone": output[6].astype(np.int)}))

      
    tds = xr.concat([df.to_xarray() for df in dfs], dim="thetas")
    tds["thetas"] = thetas

    tstep = str(self.step)
    while(len(tstep)<6):
      tstep = "0" +  tstep

    tds.to_zarr(self.output_prefix + "%s_mainDDinfo.zarr"%(tstep))

    return thetas

  @ksnSA_cross_theta.compute
  def _ksnSA_cross_theta(self):

    dfs = []
    
    thetas = np.arange(0.05,1,0.05)
    dfs = []

    #Getting the basin in Single Flow direction
    basins = fhl.basination_SF(self.fs_context["stack"].astype('int') - 1, self.fs_context["rec"].astype('int') - 1).ravel()
    basin_size = []

    basin_median_X = []
    basin_1stQ_X = []
    basin_3rdQ_X = []

    basin_median_Y = []
    basin_1stQ_Y = []
    basin_3rdQ_Y = []

    basin_median_ksn = []
    basin_1stQ_ksn = []
    basin_3rdQ_ksn = []

    basin_median_DA = []
    basin_1stQ_DA = []
    basin_3rdQ_DA = []

    basin_median_E = []
    basin_1stQ_E = []
    basin_3rdQ_E = []

    nodeids = np.arange(self.nx * self.ny)

    ksnftheta = {}
    for theta in thetas:
      tksn = np.power(self.flowacc.ravel(),theta) * self.get_slope()
      ksnftheta[theta] = tksn

    nbas = 0
    for bs in np.unique(basins):

      mask = (basins == bs) & (self.flowacc.ravel() > self.minAcc)
      if(mask[mask==True].shape[0] == 0):
        continue
      nbas+=1
      basin_size.append(self.flowacc.ravel()[mask].max())
      row,col = self.iteratool.node2rowcol(nodeids[mask])
      X = col * self.nx + self.nx/2
      Y = row * self.ny + self.ny/2
      basin_median_X.append(np.median(X))
      basin_1stQ_X.append(np.percentile(X,25))
      basin_3rdQ_X.append(np.percentile(X,75))
      basin_median_Y.append(np.median(Y))
      basin_1stQ_Y.append(np.percentile(Y,25))
      basin_3rdQ_Y.append(np.percentile(Y,75))
      basin_median_DA.append(np.median(self.flowacc.ravel()[mask]))
      basin_1stQ_DA.append(np.percentile(self.flowacc.ravel()[mask], 25))
      basin_3rdQ_DA.append(np.percentile(self.flowacc.ravel()[mask], 75))
      basin_median_E.append(np.median(self.erosion_rate.ravel()[mask]))
      basin_1stQ_E.append(np.percentile(self.erosion_rate.ravel()[mask], 25))
      basin_3rdQ_E.append(np.percentile(self.erosion_rate.ravel()[mask], 75))

      tbasin_median_ksn = []
      tbasin_1stQ_ksn = []
      tbasin_3rdQ_ksn = []
      for theta in thetas:
        tksn = ksnftheta[theta]
        tbasin_median_ksn.append(np.median(tksn[mask]))
        tbasin_1stQ_ksn.append(np.percentile(tksn[mask], 25))
        tbasin_3rdQ_ksn.append(np.percentile(tksn[mask], 75))

      basin_median_ksn.append(tbasin_median_ksn)
      basin_1stQ_ksn.append(tbasin_1stQ_ksn)
      basin_3rdQ_ksn.append(tbasin_3rdQ_ksn)
      
      

      # output = fhl.chi_contrast_across_divides(main_divide, this_chi.ravel(), self.iteratool, self.main_divide_contrast_distance )
      # dfs.append(pd.DataFrame({"X": output[0], "Y": output[1], "chi_med_zone": output[2], "chi_med_others": output[3],
      #   "chi_max_zone": output[4], "chi_max_others": output[5], "zone": output[6].astype(np.int)}))

    # tds = xr.concat([df.to_xarray() for df in dfs], dim="thetas")
    # tds["thetas"] = thetas
    # nbas = np.unique(basins).ravel().basishape[0]
    tds = xr.Dataset({
      'n_basins': np.arange(nbas),
      'thetas': thetas,
      'basin_size': xr.DataArray(data = basin_size, dims = ('n_basins')),
      'basin_median_X': xr.DataArray(data = basin_median_X, dims = ('n_basins')),
      'basin_1stQ_X': xr.DataArray(data = basin_1stQ_X, dims = ('n_basins')),
      'basin_3rdQ_X': xr.DataArray(data = basin_3rdQ_X, dims = ('n_basins')),
      'basin_median_Y': xr.DataArray(data = basin_median_Y, dims = ('n_basins')),
      'basin_1stQ_Y': xr.DataArray(data = basin_1stQ_Y, dims = ('n_basins')),
      'basin_3rdQ_Y': xr.DataArray(data = basin_3rdQ_Y, dims = ('n_basins')),
      'basin_median_ksn': xr.DataArray(data = basin_median_ksn, dims = ('n_basins', 'thetas')),
      'basin_1stQ_ksn': xr.DataArray(data = basin_1stQ_ksn, dims = ('n_basins', 'thetas')),
      'basin_3rdQ_ksn': xr.DataArray(data = basin_3rdQ_ksn, dims = ('n_basins', 'thetas')),
      'basin_median_DA': xr.DataArray(data = basin_median_DA, dims = ('n_basins')),
      'basin_1stQ_DA': xr.DataArray(data = basin_1stQ_DA, dims = ('n_basins')),
      'basin_3rdQ_DA': xr.DataArray(data = basin_3rdQ_DA, dims = ('n_basins')),
      'basin_median_E': xr.DataArray(data = basin_median_E, dims = ('n_basins')),
      'basin_1stQ_E': xr.DataArray(data = basin_1stQ_E, dims = ('n_basins')),
      'basin_3rdQ_E': xr.DataArray(data = basin_3rdQ_E, dims = ('n_basins'))
    })

    tstep = str(self.step)
    while(len(tstep)<6):
      tstep = "0" +  tstep

    tds.to_zarr(self.output_prefix + "%s_ksn_per_basin_info.zarr"%(tstep))

    return thetas




"""Top-level package for fastscape_litho."""

__author__ = """Boris Gailleton"""
__email__ = 'boris.gailleton@gfz-potsdam.de'
__version__ = '0.0.1'


from .fastscape_litho import *
from .TA import *
from ._looper import *
from ._helper import *



"""
This module contains the xarray-simlab processes related to the 3D cube of lithology.
B.G.
"""
import numpy as np
import numba as nb
import xarray as xr
import xsimlab as xs
import pandas as pd
import math
import fastscape
from fastscape.processes.erosion import TotalErosion
from fastscape.processes.channel import DifferentialStreamPowerChannelTD
from fastscape.processes.hillslope import DifferentialLinearDiffusion
from fastscape.processes.grid import UniformRectilinearGrid2D
import xsimlab as xs

from fastscape.processes.boundary import BorderBoundary
from fastscape.processes.channel import (StreamPowerChannel,
                                 DifferentialStreamPowerChannelTD)
from fastscape.processes.context import FastscapelibContext
from fastscape.processes.flow import DrainageArea, SingleFlowRouter, MultipleFlowRouter, FlowRouter, FlowAccumulator
from fastscape.processes.erosion import TotalErosion
from fastscape.processes.grid import RasterGrid2D
from fastscape.processes.hillslope import LinearDiffusion, DifferentialLinearDiffusion
from fastscape.processes.initial import (BareRockSurface,
                                 Escarpment,
                                 FlatSurface,
                                 NoErosionHistory)
from fastscape.processes.main import (Bedrock,
                              StratigraphicHorizons,
                              SurfaceTopography,
                              SurfaceToErode,
                              TerrainDerivatives,
                              TotalVerticalMotion,
                              UniformSedimentLayer)
from fastscape.processes.marine import MarineSedimentTransport, Sea
from fastscape.processes.tectonics import (BlockUplift,
                                   SurfaceAfterTectonics,
                                   TectonicForcing,
                                   TwoBlocksUplift)
from ._looper import iterasator
import fastscape_litho._helper as fhl



##############################################################
##############################################################
####################### LABEL 3D #############################
##############################################################
##############################################################

@nb.njit()
def calculate_indices(cumulative_height, z, dz, nz, oz, indices, labelmatrix ):
  '''
    Helper function calculating the intercept indices from the 3D matrix functio of the depth
  '''
  # Iterating through the rows (i) and col (j)
  for i in range(indices.shape[0]):
    for j in range(indices.shape[1]):

      this_id = math.ceil((cumulative_height[i,j] - oz)/dz)
      if(this_id >= nz):
        indices[i,j] = -1
      else:
        indices[i,j] = labelmatrix[i,j,this_id]




@xs.process
class Label3D:
  '''
    Base class for the 3D label, it only manages the geometry of it: ingesting a 3D matrix of indices,
    and managing the intersection with the topography and erosion.
    Processes with variable characteristics need to be registered separatedly.
    Created: 03/2021
    Last Mod.: 03/2021
    Authors: B.G.  
  '''

  # First defining the variable belonging to Label3D
  dz = xs.variable(
    default = 50, 
    description = 'Depth resolution of the 3D matrix of labels'
  )

  origin_z = xs.variable(
    default = 0,
    description = 'original elevation of the grid. Leave to 0'
  )

  z = xs.index(
    dims='z',
    description='Matrix depth coordinate'
  )

  labelmatrix = xs.variable(
    dims = ('y','x','z'),
    description = "The 3D numpy array of indices. dimensions are row, col, depth indice."
  )

  indices = xs.variable(
    intent = 'out',
    dims=[(), ('y', 'x')],
    description= 'Indices of the depth matrix intersection at runtime, it is used to get the 2D arrays of rock-dependent characteristics.'
  )


  # Some foreign variables needed for the computing
  ## Cumulative height is the cumulative erosion into bedrock 
  cumulative_height = xs.foreign(TotalErosion, "cumulative_height")
  ## Shape is the shape of the 2D array of topography
  shape = xs.foreign(UniformRectilinearGrid2D,"shape")


  def initialize(self):
    # initialising the depth geometry
    self.nz = self.labelmatrix.shape[2]
    self.z = np.linspace(-1 * self.origin_z, -1 * self.origin_z + self.nz * self.dz + self.dz , self.nz)
    # Adn the indices
    self.indices = np.full(self.shape, -1, dtype = np.int32)


  def run_step(self):
    #At the start, there is no array so the original depth is 0
    if(isinstance(self.cumulative_height,np.ndarray) ==  False):
      self.indices = np.zeros(self.shape, dtype = np.int32)
    else:
      # Else I am calculating the intercept calling the numba function above
      calculate_indices(self.cumulative_height, self.z, self.dz, self.nz, self.origin_z, self.indices, self.labelmatrix )



##############################################################
##############################################################
########### Adaptation of fastscape processes ################
##############################################################
##############################################################


@xs.process
class DifferentialStreamPowerChannelTDForeign(DifferentialStreamPowerChannelTD):
  k_coef_bedrock = xs.variable(intent = 'out',
    dims=[(), ('y', 'x')],
    description='bedrock channel incision coefficient'
  )
  Kr_lab = xs.variable(dims = 'n_labels', description = 'Kr value for each label')
  indices = xs.foreign(Label3D, "indices")

  def run_step(self):
    self.k_coef_bedrock = self.Kr_lab[self.indices]
    super(DifferentialStreamPowerChannelTDForeign, self).run_step()


@xs.process
class DifferentialLinearDiffusionForeign(DifferentialLinearDiffusion):
  diffusivity_bedrock = xs.variable(
    intent = 'out',
    dims=[(), ('y', 'x')],
    description='bedrock diffusivity'
  )
  diffusivity_soil = xs.variable(
    intent = 'out',
    dims=[(), ('y', 'x')],
    description='soil (sediment) diffusivity'
  )

  Kdr_lab = xs.variable(dims = 'n_labels', description = 'Kdr value for each label')
  Kds_lab = xs.variable(dims = 'n_labels', description = 'Kds value for each label')

  indices = xs.foreign(Label3D, "indices")

  def run_step(self):
    self.diffusivity_bedrock = self.Kdr_lab[self.indices]
    self.diffusivity_soil = self.Kds_lab[self.indices]
    super(DifferentialLinearDiffusionForeign, self).run_step()

@xs.process
class StreamPowerChannelForeign(StreamPowerChannel):
  k_lab = xs.variable(dims = 'n_labels', description = 'K value for each label')
  k_coef = xs.variable(intent = 'out',
    dims=[(), ('y', 'x')],
    description='bedrock channel incision coefficient'
  )
  indices = xs.foreign(Label3D, "indices")

  def run_step(self):
    self.k_coef = self.k_lab[self.indices]
    super(StreamPowerChannelForeign, self).run_step()

@xs.process
class FlowAccumulatorForeign(FlowAccumulator):
  """Accumulate the flow from upstream to downstream."""

  runoff = xs.variable(
    intent = 'out',
    dims=[(), ('y', 'x')],
    description='surface runoff (source term) per area unit'
  )

  precipitations = xs.variable(
    intent = 'in',
    dims=[(), ('y', 'x')],
    description='surface runoff (source term) per area unit',
    default = 1
  )

  infiltration_lab = xs.variable(dims = 'n_labels', description = 'infiltration values for each label')

  shape = xs.foreign(UniformRectilinearGrid2D, 'shape')
  cell_area = xs.foreign(UniformRectilinearGrid2D, 'cell_area')
  stack = xs.foreign(FlowRouter, 'stack')
  nb_receivers = xs.foreign(FlowRouter, 'nb_receivers')
  receivers = xs.foreign(FlowRouter, 'receivers')
  weights = xs.foreign(FlowRouter, 'weights')
  indices = xs.foreign(Label3D, "indices")


  flowacc = xs.variable(
      dims=('y', 'x'),
      intent='out',
      description='flow accumulation from up to downstream'
  )

  def run_step(self):
    self.runoff = self.precipitations - self.infiltration_lab[self.indices]

    field = (self.runoff * self.cell_area).flatten()
    # print("here1")
    if self.receivers.ndim == 1:
      fhl._flow_accumulate_sd(field, self.stack, self.receivers)

    else:
      fhl._flow_accumulate_mfd(field, self.stack, self.nb_receivers,
                             self.receivers, self.weights)

    self.flowacc = field.reshape(self.shape)
    self.flowacc[self.flowacc <= 0] = 1
    # print("here2:", self.flowacc)


@xs.process
class FlexureLabel(fastscape.processes.Flexure):
  """Flexural isostatic effect of both erosion and tectonic
  forcing.

  """
  lithos_density = xs.variable(
      dims=[(), ('y', 'x')],
      description='lithospheric rock density (in this case, given)',
      intent = 'out'
  )

  rho_lab = xs.variable(dims = 'n_labels', description = 'Density value for each label')

  indices = xs.foreign(Label3D, "indices")


  def run_step(self):

    self.lithos_density = self.rho_lab[self.indices]
    super(FlexureLabel, self).run_step()


  # UniformRectilinearGrid2D

##############################################################
##############################################################
####################### Models ###############################
##############################################################
##############################################################



# A model with 3D litho block and diffusion/sediment stream power
sediment_model_label3D = fastscape.models.basic_model.update_processes({
    'bedrock': Bedrock,
    'active_layer': UniformSedimentLayer,
    'init_bedrock': BareRockSurface,
    'flow': MultipleFlowRouter,
    'spl': DifferentialStreamPowerChannelTDForeign,
    'diffusion': DifferentialLinearDiffusionForeign,
    'label': Label3D,
    'drainage': DrainageArea
})

#A fast, SPL single flow only model for quick tests 
fast_SPL_model_label3D = fastscape.models.basic_model.update_processes({
    'bedrock': Bedrock,
    'active_layer': UniformSedimentLayer,
    'init_bedrock': BareRockSurface,
    'flow': SingleFlowRouter,
    'spl': StreamPowerChannelForeign,
    'label': Label3D,
    'drainage': DrainageArea
}).drop_processes({'diffusion'})

# A basic set up for 3D block without any surface process activated (can be used to create new models)
basic_model_label3D = fast_SPL_model_label3D.drop_processes(['spl'])

# A model utilising the full power of the 3D block possibilities
full_model_label3D = fastscape.models.basic_model.update_processes({
    'bedrock': Bedrock,
    'active_layer': UniformSedimentLayer,
    'init_bedrock': BareRockSurface,
    'flow': MultipleFlowRouter,
    'spl': DifferentialStreamPowerChannelTDForeign,
    'diffusion': DifferentialLinearDiffusionForeign,
    'label': Label3D,
    'drainage': DrainageArea
})














































# End of file


"""
TAKEN FROM lsdnumbatools, temporary measure before the latter is available
this provides a numba jitclass to loop through node effiviently

B.G.
"""
import numpy as np
import numba as nb
import math as m
from numba.experimental import jitclass

@nb.njit(inline='always')
def _get_neighbouring_nodes_with_checks(output,adder,node):
	i = 0
	for add in adder:
		if(add != -2147483648):
			output[i] = node + add
		else:
			output[i] = -2147483648
		i += 1


# specs for the jitclass which needs to be statically typed
spec = [
		('node_type', nb.int16[:]),
		('nx', nb.int32),
		('ny', nb.int32),
		('nxy', nb.int32),
		('dx', nb.float64),
		('dy', nb.float64),
		('dxy', nb.float64),
		('nodata', nb.int32),
		('neighbourer', nb.int32[:,:]),
		('lengthener', nb.float64[:]),
	]
	#min, max int32 = -2,147,483,648 2,147,483,647

@jitclass(spec)
class iterasator(object):
	"""
		The iterasator class is made to return neighboring informations from a node.
		It takes care of boundary conditions for example, where the geometric neighbors are not the right ones.
		Date: November 2020
		Authors: B.G.
	"""
	def __init__(self,node_type,nx,ny, dx, dy):
		"""
		Initialises the iterasator. 
		node_type (1D numpy array of int): vectorised array of node type
		nx (int): number of cols
		ny (int): number of rows
		Date: November 2020
		Authors: B.G.
		"""
		# The nodes type are documented in the file node_graph.py
		self.node_type = node_type
		# Number of cols
		self.nx = nx
		# Number of rows
		self.ny = ny

		#number of elements
		self.nxy = nx*ny

		# No data value for the neighbourer
		self.nodata = -2147483648

		self.dx = dx
		self.dy = dy
		self.dxy = m.sqrt(dx**2 + dy**2)
		# The neighbourer is the most important (and irksome to code) structure of this class:
		# It hardcodes the window looping through vectorised neighbours for each cases
		# I adopted this approach rather than padding no data to be as flexible as possible
		self.neighbourer = np.array([
[ - nx - 1, - nx, - nx + 1, -1, 1, nx - 1, nx, nx + 1], # neighbourer 0: classic internal D8 node and all the neighbors are returned
[ - 1, - nx, - nx + 1, nx - 1, 1, 2 * nx - 1, nx, nx + 1], # neighbourer 1: periodic D8 node at the Western Boundary
[ - 1 + ny * nx,  ny * nx, ny * nx + 1, -1, 1, nx - 1, nx, nx + 1], # neighbourer 2: periodic D8 node at the Northern Boundary
[ - nx - 1, - nx, - 2 * nx + 1, -1, - nx + 1, nx - 1, nx, 1], # neighbourer 3: periodic D8 node at the Eastern Boundary
[ - nx - 1, - nx, - nx + 1, -1, 1, - ny * nx - 1, - ny * nx, - ny * nx + 1], # neighbourer 4: periodic D8 node at the Southern Boundary
[ self.nodata, self.nodata, self.nodata, -1, 1, nx - 1, nx, nx + 1], # neighbourer 5: classic D8 node Northern boundary
[ self.nodata, - nx, - nx + 1, self.nodata, 1, self.nodata, nx, nx + 1], # neighbourer 6: classic D8 node Western boundary
[ - nx - 1, - nx, - nx + 1, -1, 1, self.nodata, self.nodata, self.nodata], # neighbourer 7: classic D8 node Southern boundary
[ - nx - 1, - nx, self.nodata, -1, self.nodata, nx - 1, nx, self.nodata], # neighbourer 8: classic D8 node Eastern boundary
[ self.nodata,  self.nodata, self.nodata, self.nodata, 1, self.nodata, nx, nx + 1], # neighbourer 9: classic D8 node at the Northern-western corner Boundary
[ self.nodata, - nx, - nx + 1, self.nodata, 1, self.nodata, self.nodata, self.nodata], # neighbourer 10: classic D8 node at the Southwestern corner Boundary
[ - nx - 1, - nx, self.nodata, -1, self.nodata, self.nodata, self.nodata, self.nodata], # neighbourer 11: classic D8 node at the Southeastern corner Boundary
[ self.nodata, self.nodata, self.nodata, -1, self.nodata, nx - 1, nx, self.nodata], # neighbourer 12: classic D8 node at the Northeastern corner Boundary
# To add for the future, not urgent: 
# - Cyclic boundaries cause why not
# - specific corner boundary (e.g. NW with W periodic and NW with N periodic)
# - Internal periodic? (i.e. goes straight to the next periodic in the interested direction)
		],dtype = nb.int32)

		# The lengthener is associated with the neighbourer to return the length to the neighbour associated to each one
		self.lengthener = np.array([self.dxy,self.dy,self.dxy,self.dx,self.dx,self.dxy,self.dy,self.dxy], nb.float64)

	def node2rowcol(self,node):
		"""
		Converts vectorised node ID to row index and col index
		"""
		r = node // self.nx
		c = node % self.nx
		return r,c




	def get_neighbouring_nodes(self,node):
		"""
		return all the neighboring nodes of a diven node index
		"""
		output = np.full(8,self.nodata,dtype = nb.int32)
		# case 1: active, classic internal node, keeping it test-free to optimise code as this will be the overwhelming majority of cases
		# THE ONLY REQUIREMENT IS THAT NO BOUNDARY NODES ARE LABELLED WITH 1
		if(self.node_type[node] == 1):
			i = 0
			for add in self.neighbourer[0]:
				output[i] = node + add
				i += 1
		# North row (in array point of view)
		elif(node < self.nx):
			# classic border bourndary
			if(self.node_type[node] == 0):
				# N case
				if(node > 0 and node < self.nx - 1):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[5],node)
				# NW case
				elif (node == 0):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[9],node)
				# NE case
				elif (node == self.nx - 1):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[12],node)
			# periodic boundaries
			if(self.node_type[node] == 2):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[2],node)
		# South row 
		elif(node >= (self.nx * self.ny) - self.nx):
			# classic border bourndary
			if(self.node_type[node] == 0):
				# Middle of the S row
				if(node > (self.nx * self.ny - self.nx) and node < (self.nx * self.ny - self.nx) - 1):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[7],node)
				# SW classic boulder
				elif (node == (self.nx * self.ny - self.nx)):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[10],node)
					# SE classic boundary
				elif (node == (self.nx * self.ny - self.nx) - 1):
					_get_neighbouring_nodes_with_checks(output,self.neighbourer[11],node)
			elif(self.node_type[node] == 2):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[4],node)
	    # first Column
		elif(node % self.nx == 0):
			if(self.node_type[node] == 0):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[6],node)
			elif(self.node_type[node] == 2):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[1],node)
	    # Last Column
		elif(node % self.nx - 1 == 0):
			if(self.node_type[node] == 0):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[8],node)
			elif(self.node_type[node] == 2):
				_get_neighbouring_nodes_with_checks(output,self.neighbourer[3],node)

		return output



























#End of fucking file (I am extremely frustrated at Sublime text which delete the last empty rows of a file automatically and wants you to then work using obly the absolute bottom of your screen. I am not alone right?)



import numpy as np
import numba as nb
from ._looper import iterasator


@nb.njit
def _flow_accumulate_sd(field, stack, receivers):
  for inode in stack[-1::-1]:
    if receivers[inode] != inode:
      field[receivers[inode]] += field[inode]
    if(field[inode] < 0):
      field[inode] = 0


@nb.njit
def _flow_accumulate_mfd(field, stack, nb_receivers, receivers, weights):
  for inode in stack:
    if nb_receivers[inode] == 1 and receivers[inode, 0] == inode:
      continue

    for k in range(nb_receivers[inode]):
      irec = receivers[inode, k]
      field[irec] += field[inode] * weights[inode, k]
    if(field[inode] < 0):
      field[inode] = 0

@nb.njit()
def chiculation_SF(stack, receivers, nb_donors, donors, lengths, elevation, area, A0, theta, minAcc):

  # First I am getting the length to donors
  # length2donors = np.full_like(donors,-1)
  # for inode in stack:
  #   for idon in range(nb_donors[inode]):
  #     length2donors[inode,idon] = lengths[donors[inode,idon]]

  chi = np.zeros_like(stack, dtype = np.float64)
  for inode in stack:
    if(inode == receivers[inode] or area[inode] < minAcc):
      continue

    # chi[inode] = chi[receivers[inode]] + lengths[inode]/2 * ( ((A0/area[inode]) ** theta)  + ((A0/area[receivers[inode]]) ** theta) ) 
    chi[inode] = chi[receivers[inode]] + lengths[inode] * ( ((A0/area[inode]) ** theta)) 

  return chi


@nb.njit()
def chiculation_MF(mstack, receivers, nb_receivers, lengths, weights, elevation, area, A0, theta, minAcc):


  chi = np.zeros_like(mstack, dtype = np.float64)
  for inode in mstack[::-1]:

    if(nb_receivers[inode] == 0 or area[inode] < minAcc):
      continue

    chi[inode] = 0
    for j in range(nb_receivers[inode]):
      # chi[inode] += weights[inode,j] * (chi[receivers[inode,j]] + lengths[inode,j]/2 * ( ((A0/area[inode]) ** theta)  + ((A0/area[receivers[inode,j]]) ** theta) ) )
      chi[inode] += weights[inode,j] * (chi[receivers[inode,j]] + lengths[inode,j] * ( ((A0/area[inode]) ** theta)) )

  return chi


@nb.njit()
def ksn_calculation_SF(elevation, chi, receivers, stack):
  
  ksn = np.zeros_like(chi.ravel())

  for i in range(chi.shape[0]):
    irec = receivers[i]
    if(irec == i or chi[i] == 0):
      # n_ignored += 1
      continue

    ksn[i] = elevation[i] - elevation[irec]
    ksn[i] = ksn[i]/(chi[i] - chi[irec])
    # if(ksn[i] == 0):
    #   n_0 += 1

    if(ksn[i]<0):
      ksn[i] = 0
  # print(n_ignored)

  return ksn

@nb.njit()
def ksn_calculation_MF(elevation, chi, receivers,nb_receivers, weights, stack):
  
  ksn = np.zeros_like(chi.ravel())

  for i in range(chi.shape[0]):
    
    if(chi[i] == 0):
        continue

    this_ksn = 0

    for j in range(nb_receivers[i]):
      irec = receivers[i,j]

      this_this_ksn = elevation[i] - elevation[irec]
      this_this_ksn = this_this_ksn/(chi[i] - chi[irec])
      this_ksn += weights[i,j] * this_this_ksn

      # if(ksn[i] == 0):
      #   n_0 += 1
    # print(n_ignored)
    ksn[i] = this_ksn
    if(ksn[i]<0):
      ksn[i] = 0

  return ksn


@nb.njit()
def slope_SF(elevation, receivers, lengths):
  slope = np.zeros_like(elevation)
  for i in range(receivers.shape[0]):
    if(i == receivers[i]):
      continue
    slope[i] = (elevation[i] - elevation[receivers[i]])/lengths[i]
  return slope

@nb.njit()
def slope_MF(elevation, receivers, lengths, nb_receivers, weights):
  slope = np.zeros_like(elevation)
  for i in range(receivers.shape[0]):
    if(nb_receivers[i] == 0):
      continue

    for j in range(nb_receivers[i]):
      slope[i] += weights[i,j] *  ((elevation[i] - elevation[receivers[i,j]])/lengths[i,j])
    # slope[i] = slope[i]/nb_receivers[i]
  return slope


@nb.njit()
def basination_SF(Sstack, receivers):
  basins = np.zeros_like(Sstack, dtype = np.int32)
  basin_ID = -1
  for inode in Sstack:
    if(inode == receivers[inode]):
      basin_ID = inode
    basins[inode] = basin_ID
  return basins

@nb.njit()
def is_draiange_divide_SF(basins, iterator):
  is_DD = np.zeros_like(basins, dtype = nb.boolean)
  for i in range(basins.shape[0]):
    val = basins[i]
    for j in iterator.get_neighbouring_nodes(i):
      if(j < 0 or is_DD[i] == 1):
        continue
      if(basins[j] != val):
        is_DD[i] = 1
        is_DD[j] = 1
  return is_DD


@nb.njit()
def extract_main_regions(stack, main_divide, receivers):

  # First step: labelling the thingy
  for i in stack:
    main_divide[i] = main_divide[receivers[i]]


@nb.njit()
def boundcheck(node, iterator):
  r, c = iterator.node2rowcol(node)
  if(r < 0 or r >= iterator.ny or c < 0 or c >= iterator.nx):
    return False
  else:
    return True



@nb.njit()
def chi_contrast_across_divides(main_divide, chi, iterator, distance ):
  
  # Formatting the output by index:
  # 0 -> X
  # 1 -> Y
  # 2 -> chi_median_zone
  # 3 -> chi_median_other
  # 4 -> chi_max_zone
  # 5 -> chi_max_other
  # 6 -> zone
  output = np.zeros((7,iterator.ny *iterator.nx ), dtype = nb.float64)
  # Number of drainage divide points
  n_elements = 0
  # The half window in integer size 
  tdx_d2 = round(distance/iterator.dx/2)
  tdy_d2 = round(distance/iterator.dy/2)

  # Iterating through all nodes
  for i in range(main_divide.shape[0]):
    # Value of the current zone
    val = main_divide[i]
    # Gathering chi values of the different zones
    vals_zone = []
    vals_other = []

    # getting the row and col index
    r, c = iterator.node2rowcol(i) 
    # Checker: if all the points gathered are within the same zone, I ignore the point
    keep = False

    # weeeeeeeee itarating aroud the node
    for tr in np.arange(r - tdy_d2,r + tdy_d2 + tdy_d2/2, 1, dtype = nb.int32):
      for tc in  np.arange(c - tdx_d2,c + tdx_d2 + tdx_d2/2, 1, dtype = nb.int32):
        # falling back into node index
        j = iterator.nx * tr + tc 

        #Chacking whether the node exists
        if(boundcheck(j,iterator) == False):
          continue

        #If the node is in a different zone: I gather the information
        if(main_divide[j] != val):
          keep = True
          # As well as saving chi for this area
          if(chi[j] > 0):
            vals_other.append(chi[j])
        else:
          # else just saving chi for the otehr area
          if(chi[j] > 0):
            vals_zone.append(chi[j])

    #if at least one node is in the other area and not null
    if(keep and len(vals_other)>0 and len(vals_zone) > 0):
      arrz = np.zeros(len(vals_zone))
      arro = np.zeros(len(vals_other))
      for I in range(len(vals_zone)):
        arrz[I] = vals_zone[I]
      for I in range(len(vals_other)):
        arro[I] = vals_other[I]

      # calculating the output as described above
      output[0,n_elements] = c * iterator.dx + iterator.dx/2
      output[1,n_elements] = r * iterator.dy + iterator.dy/2
      output[2,n_elements] = np.median(arrz)
      output[3,n_elements] = np.median(arro)
      output[4,n_elements] = np.max(arrz)
      output[5,n_elements] = np.max(arro)
      output[6,n_elements] = nb.float64(main_divide[i])
      n_elements += 1
  # Done, only returning the outputs needed
  return output[:,:n_elements]































  # End of file